extern crate regex;
use std::env;
use std::fs;
use std::io;
use std::collections::{BTreeMap, LinkedList};
use regex::Regex;

const STRING: &str = "STRING";
const NUMBER: &str = "NUMBER";

struct Parser{
    target: String,
    //Key is name of variable, value is value of variable.
    symbol_table: BTreeMap<String, VariableData>,
    //store numbers in front of each statement and check for duplicates
    //Also used for branching
    link: LinkedList<String>,
}

// allows cloning and allows rust to run equals on struct
#[derive(Clone, Debug, PartialEq)]
struct VariableData {
    val_type: String,
    value: String,
}

impl Parser {
    //A S-IMP program is a sequence of S-IMP statements delimited by newlines.
    fn parse(&mut self) {
        self.statement_list(&self.target.clone());

    }

    //a number followed by one or more non-newline whitespace characters followed by either an
    //assignment_statement, a branch_statement, a comment_line, an input_statement, or a print_statement
    fn statement_list(&mut self, s: &str) {
        //breaks the program into lines
        let statements = s.lines();
        for lines in statements {
            //evaluates each line as a statement
            self.statement(lines);
        }
    }

    //a statement can be either an assignment_statement, a branch_statement, a comment_line, an input_statement, or a print_statement
    fn statement(&mut self, s: &str) {
        let tokens: Vec<&str> = s.split(' ').collect();
        //Convert s from &str to String
        let line= s.to_string();
        //A line is NUMBER + STATEMENT
        //Split the NUMBER and STATEMENT
        let first = s.split_once(" ").unwrap();
        //Number is the NUMBER in front of the STATEMENT.
        let number = String::from(first.0);
        //Key is the key to store in symbol table
        let key = String::from(first.0);
        //check if NUMBER already exist in linked list
        if self.link.contains(&number){
            //Display error message.
            println!("Syntax error, two statements begin with the same NUMBER");
            panic!();
        } else {
            //Store NUMBER into linked list for check for duplicates and branching
            self.link.push_back(number);
            //Store NUMBER as key and the whole line as statement for branching
            //Ex:(010, 010 INPUT X15)
            let line_var = VariableData {
                val_type: STRING.parse().unwrap(),
                value: line,
            };
            self.symbol_table.insert(key, line_var);
        }

        //Statement = number + statement
        //Therefore token is at position 1 instead of 0.
        if tokens[1] == "INPUT" {
            self.input_statement(s);
        } else if  tokens[1] == "PRINT" {
            self.print_statement(s);
        } else if tokens[1] == "C" {
            self.comment_line();
        } else if tokens[1] == "BRANCH"{
            self.branch_statement(s);
        } else {
            self.assignment_statement(s);
        }
    }

    //If two variables have the same value, branch to the target line number
    fn branch_statement(&mut self, s: &str) {
        //Ex: 123 BRANCH X55 X44 010
        let tokens: Vec<&str> = s.split(' ').collect();
        //Check if the line is branching to itself
        let this = String::from(tokens[0]);
        let that = String::from(tokens[4]);
        //If it branches to itself
        if this == that{
            //Display error message
            println!("Death loop, cannot branch to itself.");
            panic!();
        }
        //Check if the two variables are equal
        let x = String::from(tokens[2]);
        let y = String::from(tokens[3]);
        //Get variables from the symbol table
        // and compare the two to see if they are equal
        let a = self.symbol_table.get(&x).unwrap();
        let b = self.symbol_table.get(&y).unwrap();
        let a = &a.value;
        let b = &b.value;
        //if two variables are equal
        if a == b {
            //num is NUM in front of statement
            let num = String::from(tokens[4]);
            let num_clone = num.clone();
            //Traverse the linked list
            //find the line to branch to
            let symbol_clone = self.symbol_table.clone();
            let mut link_clone: LinkedList<String> = self.link.clone();
            //if they don't match, visit next
            while link_clone.front() != Some(&num_clone){
                link_clone.pop_front();
            }
            //when they match, start executing from that line
            while link_clone.front() != None {
                //Get the entire line from the symbol table
                let line_to_execute = link_clone.front().unwrap();
                let line = symbol_clone.get(line_to_execute).unwrap();
                let line = &line.value;

                let line_tokens: Vec<&str> = line.split(' ').collect();
                //Call different function based on keyword
                if line_tokens[1] == "INPUT" {
                    self.input_statement(line);
                } else if line_tokens[1] == "PRINT" {
                    self.print_statement(line);
                } else if line_tokens[1] == "C" {
                    self.comment_line();
                } else if line_tokens[1] == "BRANCH"{
                    self.branch_statement(line);
                } else {
                    self.assignment_statement(line);
                }
                link_clone.pop_front();
            }
        }
        //do nothing if false
    }

    fn operation_assign(&mut self, arg1: &str, arg2: &str, operand: &str) -> VariableData {
        // check if both arguments are variables
        self.variable(arg1);
        self.variable(arg2);
        // check if both are initialized
        let arg = self.symbol_table.get(arg1);
        if arg == None {
            println!("Error, variable {} is not initialized", arg1);
            panic!();
        }
        let arg = self.symbol_table.get(arg2);
        if arg == None {
            println!("Error, variable {} is not initialized", arg2);
            panic!();
        }
        // get values and type
        let value1 = self.symbol_table.get(arg1).unwrap();
        let value1val = &value1.value;
        let value1type = &value1.val_type;
        let value2 = self.symbol_table.get(arg2).unwrap();
        let value2val = &value2.value;
        let value2type = &value2.val_type;
        // check if both arguments are numbers
        return if value1type == NUMBER && value2type == NUMBER {
            // convert to integer
            let value1int = value1val.parse::<i32>().unwrap();
            let value2int = value2val.parse::<i32>().unwrap();
            // if plus operation
            if operand == "+" {
                let result = value1int + value2int;
                // save the result (return is implicit)
                VariableData{
                    val_type: NUMBER.parse().unwrap(),
                    value: result.to_string()
                }
            } else { // otherwise minus
                let mut result = value1int - value2int;
                if result < 0 {
                    result = 0;
                }
                // save the result (return is implicit)
                VariableData {
                    val_type: NUMBER.parse().unwrap(),
                    value: result.to_string()
                }
            }
        } else if value1type == NUMBER && value2type == STRING {
            // if first argument is number and second is string
            println!("SEMANTIC ERROR: Not expecting NUMBER +/- STRING \n");
            panic!();
        } else if value1type == STRING && value2type == NUMBER {
            //if first argument is string and second is number
            let value2int = value2val.parse::<usize>().unwrap();
            // Ex: Xm is "Hello World" and Xn is 1, the Xm + Xn is "ello World".
            if operand == "+" {
                // skip first characters based on value2int
                let result:String = value1val.chars().skip(value2int).collect();
                VariableData {
                    val_type: STRING.parse().unwrap(),
                    value: result,
                }
            } else {
                // only show the number of characters designated by value2int
                // if Xm is "Hello World" and Xn is 1, then Xm - Xn is "H".
                let result_array = value1val.chars();
                let mut count = 0;
                let mut result:String = "".to_string();
                // step through each character until the number of characters processed equals Xn (second argument)
                for char in result_array {
                    if count < value2int {
                        result = format!("{}{}", result, char);
                    }
                    count = count + 1;
                }
                VariableData {
                    val_type: STRING.parse().unwrap(),
                    value: result.to_string()
                }
            }
        } else {
            // if both argument are string
            if operand == "+" {
                // concatenate string using format
                let result = format!("{}{}", value1val, value2val);
                VariableData {
                    val_type: STRING.parse().unwrap(),
                    value: result,
                }
            } else {
                // remove the strings that are similar to both arguments
                // following line splits the string using argument 2 as delimiter
                let result_array: Vec<&str> = value1val.split(value2val).collect();
                let mut result= "".to_string();
                // put all the strings together and return result
                for section in result_array {
                    result = format!("{}{}", result, section);
                }
                VariableData {
                    val_type: STRING.parse().unwrap(),
                    value: result,
                }
            }
        }
    }

    fn assignment_statement(&mut self, s: &str) {
        // Ex: 123 X4 = "hello world\n"
        //first = number, second = variable, third = '='
        let first = s.split_once(" ").unwrap();
        //Now first.0 = 123, first.1 = 'X4 = "hello world\n"'
        let second = first.1.split_once(" ").unwrap();
        //Now second.0 = X4, second.1 = '= "hello world\n"'
        //second.0 is the variable to be stored.
        let var =  String::from(second.0);
        self.variable(&var);
        let third = second.1.split_once(" ").unwrap();
        //Now third.0 = '=', third.1 = '"hello world\n"'
        if third.1.contains(" + ") {
            //Check Condition: third.1 = 'X1 + X2'
            let third_mod: Vec<&str> = third.1.split(" + ").collect();
            if third_mod.len() == 2 {
                let result = self.operation_assign(third_mod[0], third_mod[1], "+");
                self.symbol_table.insert(var, result);
            }
        } else if third.1.contains(" - ") {
            // Check condition: third.1 = 'X1 - X2'
            let third_mod: Vec<&str> = third.1.split(" - ").collect();
            if third_mod.len() == 2 {
                let result = self.operation_assign(third_mod[0], third_mod[1], "-");
                self.symbol_table.insert(var, result);
            }
        }
        else {
            //Now third.0 = '=', third.1 = '"hello world\n"' || third_mod[0] = ISSET | NUMBER | STRING
            // first check for ISSET NUMBER STRING
            let fourth: Vec<&str> = third.1.split(" ").collect();
            if fourth[0] == "ISSET" {
                // check if the variable has been used before
                // check if variable is valid
                self.variable(fourth[1]);
                // get the value struct
                let arg = self.symbol_table.get(fourth[1]);
                let mut result_int = 1;
                // if value struct is none, return 0
                // otherwise return 1
                if arg == None {
                    result_int = 0;
                }
                let result = VariableData {
                    val_type: NUMBER.parse().unwrap(),
                    value: result_int.to_string()
                };
                self.symbol_table.insert(var, result);
            } else if fourth[0] == "NUMBER" {
                // attempts to convert from string to number
                // check if variable is valid
                self.variable(fourth[1]);
                // get value struct
                let arg = self.symbol_table.get(fourth[1]).unwrap();
                let arg_value = &arg.value;
                // create result assuming that it will not be a number
                let mut result = VariableData {
                    val_type: NUMBER.parse().unwrap(),
                    value: "0".to_string()
                };
                let mut is_int:bool = true;
                // check each character if it is a numerical value
                for c in arg_value.chars() {
                    if !c.is_numeric() {
                        is_int = false;
                    }
                }
                // all are numerical set result to the argument value
                // otherwise keep it at 0
                if is_int {
                    result = VariableData {
                        val_type: NUMBER.parse().unwrap(),
                        value: arg_value.to_string()
                    };
                    self.symbol_table.insert(var, result);
                } else {
                    self.symbol_table.insert(var, result);
                }
            } else if fourth[0] == "STRING" {
                // convert into string and save
                self.variable(fourth[1]);
                let arg = self.symbol_table.get(fourth[1]).unwrap();
                let arg_value = &arg.value;
                // since argument value is already saved as a string, just change the actual value type in the struct to STRING
                // and return it
                let result = VariableData {
                    val_type: STRING.parse().unwrap(),
                    value: arg_value.to_string()
                };
                self.symbol_table.insert(var, result);
            } else {
                //Now third.0 = '=', third.1 = 1
                let third_mod: Vec<&str> = third.1.split("\"").collect();
                // check for quotes
                if third_mod.len() > 1 {
                    //Now third_mod[1] = 'hello world\n'
                    // Strip "\n" if it exists
                    let third_mod: Vec<&str> = third_mod[1].split("\\n").collect();
                    let result = VariableData {
                        val_type: STRING.parse().unwrap(),
                        value: third_mod[0].parse().unwrap(),
                    };
                    self.symbol_table.insert(var, result);
                } else {
                    //Now third_mod[0] = 1 || third_mod[0] = ISSET | NUMBER | STRING
                    let value = third_mod[0];
                    let mut is_int:bool = true;
                    for c in value.chars() {
                        if !c.is_numeric() {
                            is_int = false;
                        }
                    }
                    // assuming X11 = 12
                    let mut result = VariableData {
                        val_type: NUMBER.parse().unwrap(),
                        value: third_mod[0].parse().unwrap(),
                    };
                    // if we do not get a numerical value but a variable in arugment lik X11 = X12
                    if Regex::new(r"^X[\d]*$").unwrap().is_match(value) {
                        // it a variable copy it
                        let arg = self.symbol_table.get(third_mod[0]).unwrap();
                        let arg_value = &arg.value;
                        let arg_type = &arg.val_type;
                        result = VariableData {
                            val_type: arg_type.parse().unwrap(),
                            value: arg_value.parse().unwrap(),
                        };
                    } else if !is_int { // otherwise error message and panic
                        println!("Syntax Error: All strings should be in quotes (\").");
                        panic!();
                    }
                    // save
                    self.symbol_table.insert(var, result);
                }
            }
        }
    }

    //Comment lines does nothing
    fn comment_line(&mut self) {
        //empty
    }

    //Read string from stdin and input it into variable
    fn input_statement(&mut self, s: &str) {
        //Ex: 123 INPUT X55
        //first = number, second = INPUT, third = variable
        let first = s.split_once(" ").unwrap();
        //Now first.0 = 123, first.1 = INPUT X55
        let second = first.1.split_once(" ").unwrap();
        //Now second.0 = INPUT, second.1 = X55
        //second.1 is the variable to be stored.
        let var =  String::from(second.1);
        self.variable(&var);
        //Prompt user to enter a string
        println!("Encountered statement: {}", &s);
        println!("Please enter a string to be stored into variable {}:", &var);
        let mut input_str = String::new();
        io::stdin().read_line(&mut input_str).
            expect("Couldn't read line");
        //Insert variable and string into symbol table
        // first convert string into variableData format
        let result = VariableData {
            val_type: STRING.parse().unwrap(),
            value: input_str,
        };
        self.symbol_table.insert(var, result);
    }

    //Print either the value or the value type of the variable
    fn print_statement(&mut self, s:&str) {
        //Ex: 123 PRINT X55 VALUE || 123 PRINT X55 TYPE
        //first = number, second = PRINT, third = VALUE/TYPE
        let first = s.split_once(" ").unwrap();
        //Now first.0 = 123, first.1 = PRINT X55 VALUE/TYPE
        let second = first.1.split_once(" ").unwrap();
        //Now second.0 = PRINT, second.1 = X55 VALUE/TYPE
        let tokens: Vec<&str> = second.1.split(' ').collect();
        //Check if the statement is complete.
        if tokens.len() > 1{
            //tokens[0] of X55 VALUE/TYPE is X55
            let var = String::from(tokens[0]);
            self.variable(&var);
            //check if variable is initialized in the symbol table
            if self.symbol_table.get(&var) == None {
                println!("Error, variable {} is not initialized", var);
                panic!();
            }
            //check for keyword VALUE and type
            if tokens[1] == "VALUE"{
                let value = self.symbol_table.get(var.as_str()).unwrap();
                let result = &value.value;
                println!("The value of variable {} is: {}", tokens[0], result);
            } else if tokens[1] == "TYPE"{
                let value = self.symbol_table.get(var.as_str()).unwrap();
                let result = &value.val_type;
                println!("The type of variable {} is: {}", tokens[0], result);
            } else {
                //error when keyword isn't VALUE/TYPE
                println!("Error, invalid keyword: {}", tokens[1]);
                panic!();
            }
        } else{
            //Display error message
            println!("Error, expected VALUE/TYPE after variable");
            panic!();
        }
    }

    //a variable is X?[0-9]*
    fn variable(&mut self, s:&str) {
        if !Regex::new(r"^X[\d]*$").unwrap().is_match(s) {
            println!("invalid variable name {}", s);
            panic!();
        }
    }
}

fn main() {
    let mut args: Vec<_> = env::args().collect();
    //println!("{:?}", args);
    if args.len() > 1 {
        let filename = &args[1];
        let tgt = fs::read_to_string(filename)
            .expect("Something went wrong reading the file");
        let mut parser = Parser {
            target: tgt,
            symbol_table: BTreeMap::new(),

            link : LinkedList::new()
        };
        //Vector for keys
        let mut var =  vec!["X0".to_string(), "X1".to_string(),
                            "X2".to_string(), "X3".to_string(), "X4".to_string()];
        //Store args into the symbol table
        while !args.is_empty() {
            let arg = VariableData {
                val_type: STRING.parse().unwrap(),
                value: args.remove(0),
            };
            parser.symbol_table.insert(var.remove(0), arg);
        }
        parser.parse();
    } else {
        println!("No program");
    }
}